import numpy as np
import random
from matplotlib import path

'''
This file contains definetions of some random shapes, and provides Signed Distance Field (SDF) for
one or several of these shapes.
To use these shapes, build an new instance of SDF, then add shapes to it one by one using
add_[xxx]_to_sdf() functions. Then you get a SDF contains boolean join of these shapes.
'''


class SignedDistanceField:
    '''
    Holds grid and Signed Distance Field (SDF) data.
    Grid always starts from 0
    '''

    def __init__(self, grid_length_x, grid_length_y, grid_dim_x, grid_dim_y):
        self.grid_dim_x = grid_dim_x
        self.grid_dim_y = grid_dim_y
        self.grid_length_x = grid_length_x
        self.grid_length_y = grid_length_y
        self.grid_x, self.grid_y = np.meshgrid(
            np.linspace(0, grid_length_x, grid_dim_x),
            np.linspace(0, grid_length_y, grid_dim_y)
        )
        max_distance = np.sqrt(grid_length_x * grid_length_x + grid_length_y * grid_length_y)
        self.field = np.zeros(self.grid_x.shape) + max_distance


class RandomTransform():
    """
    Generates and holds a set of random transform parameter
    """

    def __init__(self, x_range, y_range):
        self.shrink_x = random.uniform(2, 4)
        self.shrink_y = random.uniform(2, 4)
        self.transfer_x = random.uniform(x_range[0] - 0.5, x_range[1] - 0.5)
        self.transfer_y = random.uniform(y_range[0] - 0.5, y_range[1] - 0.5)
        self.rotate_angle = random.uniform(0, 2 * np.pi)


def get_random_polygon(x_range, y_range):
    # source data
    mesh1 = np.asarray([[0.1, 0.1],
                        [0.1, 0.2],
                        [0.2, 0.3],
                        [0.5, 0.9],
                        [0.9, 0.9],
                        [0.9, 0.7],
                        [0.8, 0.6],
                        [0.9, 0.1]])
    mesh2 = np.asarray([[0.1, 0.1],
                        [0.1, 0.9],
                        [0.9, 0.9],
                        [0.9, 0.1]])
    mesh3 = np.asarray([[0.1, 0.1],
                        [0.5, 0.9],
                        [0.9, 0.1]])
    mesh4 = np.asarray([[0.1, 0.1],
                        [0.3, 0.5],
                        [0.5, 0.8],
                        [0.7, 0.5],
                        [0.9, 0.1]])

    # choose a mesh
    meshes = (mesh1, mesh2, mesh3, mesh4)
    mesh = meshes[random.randint(0, len(meshes) - 1)]

    # transform the chosen mesh randomly
    # random transform parameters
    rtp = RandomTransform(x_range, y_range)
    # do transform
    # rotate
    center_x = np.mean(mesh[:, 0])
    center_y = np.mean(mesh[:, 1])
    mesh_temp = mesh - [center_x, center_y]
    mesh = np.hstack((
        (np.cos(rtp.rotate_angle) * mesh_temp[:, 0] - np.sin(rtp.rotate_angle) * mesh_temp[:, 1])[:, np.newaxis],
        (np.cos(rtp.rotate_angle) * mesh_temp[:, 1] + np.sin(rtp.rotate_angle) * mesh_temp[:, 0])[:, np.newaxis]
    ))
    # shrink
    mesh[:, 0] = mesh[:, 0] / rtp.shrink_x
    mesh[:, 1] = mesh[:, 1] / rtp.shrink_y
    # transfer
    mesh = mesh + [center_x + rtp.transfer_x, center_y + rtp.transfer_y]

    return mesh


def generate_df_for_single_edge(grid_x, grid_y, x1, y1, x2, y2):
    '''
    Calculate distance field of a edge(i.e. a line segment)
    '''
    sdf = np.zeros(grid_x.shape)

    # vectorlized distance from point to line segment
    px = x2 - x1
    py = y2 - y1
    something = px * px + py * py

    u = ((grid_x - x1) * px + (grid_y - y1) * py) / float(something)
    u[u > 1] = 1.0
    u[u < 0] = 0.0

    x = x1 + u * px
    y = y1 + u * py
    dx = x - grid_x
    dy = y - grid_y

    return np.sqrt(dx * dx + dy * dy)


def generate_sf_for_polygon(grid_x, grid_y, polygon_array):
    '''
    Calculate sign field for a closed polygon.

    This will not check closeness of the input polygon. However, if the mesh is not closed,
    result will be inexact.
    '''

    # check input
    if not isinstance(polygon_array, np.ndarray):
        raise Exception('polygon_array is not a numpy array')
    mesh_shape = polygon_array.shape
    if not len(mesh_shape) == 2:
        raise ValueError('Expecting a 2D array, but here is a %dD array' % len(mesh_shape))
    if mesh_shape[0] < 3 or (not mesh_shape[1] == 2):
        raise ValueError('Not a valid mesh: size [%d,%d]' % mesh_shape)
    if (not isinstance(grid_x, np.ndarray)) or (not isinstance(grid_y, np.ndarray)):
        raise Exception('grid_x and grid_x shall be numpy array')

    polygon = path.Path(polygon_array)
    flags = polygon.contains_points(points=np.hstack(
        (grid_x.flatten()[:, np.newaxis], grid_y.flatten()[:, np.newaxis])))
    flags = flags.reshape(grid_x.shape)
    sf = np.ones(grid_x.shape)
    sf[flags] = -1
    return sf


def add_polygon_to_sdf(polygon_array, target_SDF):
    # check input
    if not isinstance(polygon_array, np.ndarray):
        raise Exception('polygon_array is not a numpy array')
    mesh_shape = polygon_array.shape
    if not len(mesh_shape) == 2:
        raise ValueError('Expecting a 2D array, but here is a %dD array' % len(mesh_shape))
    if mesh_shape[0] < 3 or (not mesh_shape[1] == 2):
        raise ValueError('Not a valid mesh: size [%d,%d]' % mesh_shape)
    if not isinstance(target_SDF, SignedDistanceField):
        raise Exception('input_SDF is not a SignedDistanceField object')

    # calculate SDF of the last edge as initial value
    num_edges = mesh_shape[0]
    df = generate_df_for_single_edge(target_SDF.grid_x, target_SDF.grid_y,
                                     polygon_array[num_edges - 1, 0], polygon_array[num_edges - 1, 1],
                                     polygon_array[0, 0], polygon_array[0, 1])
    # loop through other edges
    for i in range(num_edges - 1):
        df_temp = generate_df_for_single_edge(target_SDF.grid_x, target_SDF.grid_y,
                                              polygon_array[i, 0], polygon_array[i, 1],
                                              polygon_array[i + 1, 0], polygon_array[i + 1, 1])
        df = np.minimum(df, df_temp)

    # sign information
    sf = generate_sf_for_polygon(target_SDF.grid_x, target_SDF.grid_y, polygon_array)

    # join
    target_SDF.field = np.minimum(target_SDF.field, sf * df)

    return target_SDF


def add_circle_to_sdf(target_SDF, x_range, y_range):
    if not isinstance(target_SDF, SignedDistanceField):
        raise Exception('input_SDF is not a SignedDistanceField object')

    # generate a random circle
    rtp = RandomTransform(x_range, y_range)
    bound_x = np.max(target_SDF.grid_x)
    bound_y = np.max(target_SDF.grid_y)
    center_x = bound_x * (random.uniform(0.3, 0.7) + rtp.transfer_x)
    center_y = bound_y * (random.uniform(0.3, 0.7) + rtp.transfer_x)
    r = 0.35 * np.sqrt(bound_x * bound_x + bound_y * bound_y) / rtp.shrink_x

    # calculate SDF
    new_sdf = np.sqrt(np.power(target_SDF.grid_x - center_x, 2) + np.power(target_SDF.grid_y - center_y, 2)) - r
    target_SDF.field = np.minimum(target_SDF.field, new_sdf)

    return target_SDF


if __name__ == '__main__':
    sdf_builder = SignedDistanceField(1, 1, 80, 80)

    add_polygon_to_sdf(get_random_polygon(), sdf_builder)
    add_polygon_to_sdf(get_random_polygon(), sdf_builder)
    add_circle_to_sdf(sdf_builder)
    add_circle_to_sdf(sdf_builder)

    # draw
    from mpl_toolkits.mplot3d import Axes3D
    from matplotlib import cm
    import matplotlib.pyplot as plt

    fig = plt.figure()
    ax = Axes3D(fig)
    ax.plot_surface(sdf_builder.grid_x, sdf_builder.grid_y, sdf_builder.field, rstride=1, cstride=1, cmap=cm.rainbow)
    plt.show()
